//! Code for converting [Writecap]s to and from the X.509 certificate format.

use crate::{
    bterr,
    crypto::{AsymKeyPub, HashKind, RsaSsaPss, Sha2_256, Sha2_512, Sign},
    Principaled, Result, Writecap,
};
use bcder::{
    decode::{BytesSource, Constructed},
    encode::{PrimitiveContent, Values},
    BitString, Captured, Ia5String, Integer, Mode, OctetString, Oid, Tag,
};
use bytes::{BufMut, Bytes, BytesMut};
use chrono::{offset::Utc, TimeZone};
use x509_certificate::{
    asn1time::{Time, UtcTime},
    certificate::X509Certificate,
    rfc3280::Name,
    rfc5280::{
        AlgorithmIdentifier, AlgorithmParameter, Certificate, CertificateSerialNumber, Extension,
        Extensions, SubjectPublicKeyInfo, TbsCertificate, Validity, Version,
    },
};
mod private {
    use super::*;

    impl Sha2_256 {
        // The DER encoding of the OID 2.16.840.1.101.3.4.2.1
        const OID: &[u8] = &[0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01];
    }

    impl Sha2_512 {
        // The DER encoding of the OID 2.16.840.1.101.3.4.2.3
        const OID: &[u8] = &[0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03];
    }

    impl HashKind {
        const fn oid(&self) -> &'static [u8] {
            match self {
                HashKind::Sha2_256 => Sha2_256::OID,
                HashKind::Sha2_512 => Sha2_512::OID,
            }
        }
    }

    /// The DER encoding of 1.2.840.113549.1.1.8 (id-mgf1 in RFC 4055)
    const MGF1_OID: &[u8] = &[0x2A, 0x86, 0x48, 0x86, 0xF7, 0x0D, 0x01, 0x01, 0x08];

    struct NullTermOid {
        oid: Oid,
    }

    impl NullTermOid {
        fn new(oid: Oid) -> Self {
            Self { oid }
        }

        fn encode_ref(&self) -> impl Values + '_ {
            self.encode_ref_as(Tag::SEQUENCE)
        }

        fn encode_ref_as(&self, tag: Tag) -> impl Values + '_ {
            bcder::encode::sequence_as(
                tag,
                (
                    self.oid.encode_ref(),
                    bcder::encode::Constructed::new(Tag::NULL, bcder::encode::Nothing),
                ),
            )
        }
    }

    struct AlgoId {
        algorithm: Oid,
        parameters: NullTermOid,
    }

    impl AlgoId {
        fn new(hash_kind: HashKind) -> Self {
            let algorithm = Oid(Bytes::from(MGF1_OID));
            let parameters = NullTermOid::new(Oid(Bytes::from(hash_kind.oid())));
            Self {
                algorithm,
                parameters,
            }
        }

        fn encode_ref(&self) -> impl Values + '_ {
            self.encode_ref_as(Tag::SEQUENCE)
        }

        fn encode_ref_as(&self, tag: Tag) -> impl Values + '_ {
            bcder::encode::sequence_as(
                tag,
                (self.algorithm.encode_ref(), self.parameters.encode_ref()),
            )
        }
    }

    struct RsaSsaPssParams {
        hash_algorithm: NullTermOid,
        mask_gen_algorithm: AlgoId,
        salt_length: Integer,
        trailer_field: Option<Integer>,
    }

    impl RsaSsaPssParams {
        fn new(hash_kind: HashKind) -> Self {
            let hash_algorithm = NullTermOid::new(Oid(Bytes::from(hash_kind.oid())));
            let mask_gen_algorithm = AlgoId::new(hash_kind);
            let salt_length = Integer::from(hash_kind.len() as u64);
            Self {
                hash_algorithm,
                mask_gen_algorithm,
                salt_length,
                trailer_field: None,
            }
        }

        fn encode_ref(&self) -> impl bcder::encode::Values + '_ {
            self.encode_ref_as(bcder::Tag::SEQUENCE)
        }

        fn encode_ref_as(&self, tag: Tag) -> impl bcder::encode::Values + '_ {
            use bcder::encode::Constructed;
            bcder::encode::sequence_as(
                tag,
                (
                    Constructed::new(Tag::CTX_0, self.hash_algorithm.encode_ref()),
                    Constructed::new(Tag::CTX_1, self.mask_gen_algorithm.encode_ref()),
                    Constructed::new(Tag::CTX_2, self.salt_length.encode()),
                    self.trailer_field
                        .as_ref()
                        .map(|e| Constructed::new(Tag::CTX_3, e.encode())),
                ),
            )
        }
    }

    impl RsaSsaPss {
        const USE_PSS_OID: bool = false;

        /// The OID 1.2.840.113549.1.1.10 (RSASSA-PSS)
        const RSA_PSS_OID: &[u8] = &[0x2A, 0x86, 0x48, 0x86, 0xF7, 0x0D, 0x01, 0x01, 0x0A];

        /// The OID 1.2.840.113549.1.1.1 (RSA-ES)
        const RSA_ES_OID: &[u8] = &[0x2A, 0x86, 0x48, 0x86, 0xF7, 0x0D, 0x01, 0x01, 0x01];

        const fn oid(&self) -> &'static [u8] {
            if Self::USE_PSS_OID {
                Self::RSA_PSS_OID
            } else {
                Self::RSA_ES_OID
            }
        }

        fn params(&self) -> Result<Option<AlgorithmParameter>> {
            if Self::USE_PSS_OID {
                let params = RsaSsaPssParams::new(self.hash_kind);
                let values = params.encode_ref();
                let captured = Captured::from_values(bcder::Mode::Der, values);
                Ok(Some(AlgorithmParameter::from_captured(captured)))
            } else {
                Ok(None)
            }
        }
    }

    impl Sign {
        const fn oid(&self) -> &'static [u8] {
            match self {
                Sign::RsaSsaPss(inner) => inner.oid(),
            }
        }

        fn params(&self) -> Result<Option<AlgorithmParameter>> {
            match self {
                Sign::RsaSsaPss(inner) => inner.params(),
            }
        }
    }

    fn scheme_to_algo_id(scheme: &Sign) -> Result<AlgorithmIdentifier> {
        Ok(AlgorithmIdentifier {
            algorithm: oid(scheme.oid()),
            parameters: scheme.params()?,
        })
    }

    fn oid(slice: &'static [u8]) -> Oid {
        Oid(Bytes::from(slice))
    }

    macro_rules! bit_string {
        ($bytes:expr) => {
            BitString::new(0, Bytes::from($bytes))
        };
    }

    /// OID 2.5.29.17 (Subject Alternative Name)
    const SAN_OID: &[u8] = &[0x55, 0x1D, 0x11];

    struct SubjectAltName {
        issued_to: Ia5String,
    }

    impl SubjectAltName {
        fn new(issued_to: String) -> Result<Self> {
            let issued_to = Ia5String::from_string(issued_to).map_err(|err| bterr!("{:?}", err))?;
            Ok(Self { issued_to })
        }

        fn encode(&self) -> impl Values + '_ {
            self.encode_as(Tag::SEQUENCE)
        }

        fn encode_as(&self, tag: Tag) -> impl Values + '_ {
            bcder::encode::sequence_as(tag, self.issued_to.encode_ref())
        }

        fn encode_der(&self) -> Result<Bytes> {
            let mut writer = BytesMut::new().writer();
            self.encode().write_encoded(bcder::Mode::Der, &mut writer)?;
            Ok(writer.into_inner().into())
        }
    }

    impl AsymKeyPub<Sign> {
        fn subject_public_key_info(&self) -> Result<SubjectPublicKeyInfo> {
            Ok(SubjectPublicKeyInfo {
                algorithm: scheme_to_algo_id(&self.scheme)?,
                subject_public_key: self.to_bit_string()?,
            })
        }

        fn to_bit_string(&self) -> Result<BitString> {
            let der = self.pkey.public_key_to_der()?;
            let source = BytesSource::new(Bytes::from(der));
            let spki = Constructed::decode(source, Mode::Der, SubjectPublicKeyInfo::take_from)?;
            Ok(spki.subject_public_key)
        }

        pub fn to_der(&self) -> Result<Vec<u8>> {
            let spki = self.subject_public_key_info()?;
            let mut vec = Vec::new();
            spki.encode_ref().write_encoded(Mode::Der, &mut vec)?;
            Ok(vec)
        }
    }

    impl Writecap {
        pub fn to_cert_chain(
            &self,
            subject_key: &AsymKeyPub<Sign>,
        ) -> Result<Vec<X509Certificate>> {
            let mut chain = match self.next.as_ref() {
                Some(next) => next.as_ref().to_cert_chain(&self.body.signing_key)?,
                None => Vec::new(),
            };

            let version = Some(Version::V3);
            let serial_number = CertificateSerialNumber::from(1);
            let signature_algorithm = scheme_to_algo_id(&self.body.signing_key.scheme)?;
            let mut issuer = Name::default();
            issuer
                .append_common_name_utf8_string(&self.body.signing_key.principal().to_string())
                .map_err(|_| bterr!("failed to create issuer common name"))?;
            let expires = Utc
                .timestamp_millis_opt(1000 * self.body.expires.to_unix())
                .single()
                .ok_or_else(|| {
                    bterr!("failed to convert writecap expiration to chrono DataTime")
                })?;
            let validity = Validity {
                not_before: Time::UtcTime(UtcTime::now()),
                not_after: Time::from(expires),
            };
            let issued_to = self.body.issued_to.to_string();
            let mut subject = Name::default();
            subject
                .append_common_name_utf8_string(&issued_to)
                .map_err(|_| bterr!("failed to create subject common name"))?;
            let subject_public_key_info = subject_key.subject_public_key_info()?;
            let mut extensions = Extensions::default();
            let san = SubjectAltName::new(issued_to)?;
            extensions.push(Extension {
                id: oid(SAN_OID),
                critical: Some(false),
                value: OctetString::new(san.encode_der()?),
            });
            let tbs_certificate = TbsCertificate {
                version,
                serial_number,
                signature: signature_algorithm.clone(),
                issuer,
                validity,
                subject,
                subject_public_key_info,
                issuer_unique_id: None,
                subject_unique_id: None,
                extensions: Some(extensions),
                raw_data: None,
            };
            let cert = Certificate {
                tbs_certificate,
                signature_algorithm,
                signature: bit_string!(self.signature.data.clone()),
            };
            chain.push(cert.into());
            Ok(chain)
        }
    }
}

#[cfg(test)]
mod tests {
    use webpki::EndEntityCert;

    use crate::test_helpers::node_creds;

    #[allow(dead_code)]
    fn save_first_writecap_der_to_file() {
        let node_creds = node_creds();
        let chain = node_creds
            .writecap
            .as_ref()
            .unwrap()
            .to_cert_chain(&node_creds.sign.public)
            .unwrap();
        let first = chain.first().unwrap();
        let mut buf = Vec::new();

        first.encode_der_to(&mut buf).unwrap();
        std::fs::write("/tmp/cert.der", buf).unwrap();

        let pem = first.encode_pem().unwrap();
        std::fs::write("/tmp/cert.pem", &pem).unwrap();
    }

    #[test]
    fn node_writecap_to_cert_chain() {
        let node_creds = node_creds();

        let result = node_creds
            .writecap
            .as_ref()
            .unwrap()
            .to_cert_chain(&node_creds.sign.public);

        assert!(result.is_ok())
    }

    #[test]
    fn node_writecap_to_cert_chain_end_cert_can_be_parsed() {
        let node_creds = node_creds();
        let chain = node_creds
            .writecap
            .as_ref()
            .unwrap()
            .to_cert_chain(&node_creds.sign.public)
            .unwrap();
        let der = chain.first().unwrap().encode_der().unwrap();

        let result = EndEntityCert::try_from(der.as_slice());

        result.unwrap();
    }
}
